@@ -1,6 +1,6 @@
 /*
 
-Copyright 1984, 1988, 1990, 1993 by Rick Jansen, all rights reserved.
+Copyright 1984, 1988, 1990, 1993, 1994 by Rick Jansen, all rights reserved.
 
   X S N O W 
         *
@@ -16,6 +16,32 @@ Copyright 1984, 1988, 1990, 1993 by Rick Jansen, all rights reserved.
  But, little boys grow up, and when they are forced to a workstation
  they want their thingies there too. So here is Xsnow. For *you*.
 
+ Happy winter and a merry X-mas!
+
+ Rick Jansen
+ (rick@sara.nl)
+
+
+ Availability
+ ------------
+ Xsnow is available directly from me via my WWW page:
+
+   http://www.sara.nl/Rick.Jansen
+
+ Click on the appropriate item and the tar file will be downloaded
+ upon your request.
+
+ Xsnow is also archived at ftp.x.org in the contrib/games directory.
+
+ If all else fails you can get a new version from me vi e-mail, as
+ an uuencoded and compressed tar-file. (email: rick@sara.nl)
+
+ Copyright
+ ---------
+ Xsnow is available freely and you may give it to other people as is, 
+ but I retain all rights. Therefore it does not classify as 'Public 
+ Domain' software.
+
  Credits
  -------
  Xsnow borrows code for drawing on the desktop (the 'root' window as
@@ -30,12 +56,14 @@ Copyright 1984, 1988, 1990, 1993 by Rick Jansen, all rights reserved.
  The complete distribution of vroot is available in the 'contrib'
  directory of 'ftp.x.org' via anonymous ftp as 'vroot.shar.Z' 
 
- The big Santa with the nice moving antlers and reins  was made by 
+ The big Santa with the nice moving antlers and reins was made by 
  Holger Veit (Holger.Veit@gmd.de).
-    
- Xsnow may be freely distributed provided that all copyright notices 
- are retained.
 
+ The code for wind was done by Eiichi TAZOE (tazoe@yamato.ibm.co.jp,
+ tazoe@vnet.ibm.com)
+    
+ Installation
+ ------------
   To build:
       xmkmf
       make depend
@@ -43,10 +71,10 @@ Copyright 1984, 1988, 1990, 1993 by Rick Jansen, all rights reserved.
   or:
       cc -o xsnow snow.c -lX11 [-lsocketorwhatever] [-lm] [-l...]
 
- Happy winter and a merry X-mass!
-
- Rick Jansen
- (rick@sara.nl)
+ Silicon Graphics
+ ----------------
+ If you run Irix 5.x with virtual desks and all, you may not see 
+ any snow at all. See the README file for a solution.
 
  Modification history
  03DEC93 version 0
@@ -68,12 +96,39 @@ Copyright 1984, 1988, 1990, 1993 by Rick Jansen, all rights reserved.
  21DEC93 In case of negative x coordinates adapt width (CalcWindowTops)
  22DEC93 Rudolf's red nose. Silly. Grmbll.
  22DEC93 Bug with -nokeepsnowonwindows removed
+ 10JAN94 Improved handling of erasing snow at screen bottom in case of 
+         window in or under the snow.
+ 18FEB94 X-speed really 0 inside snow layers - prevents snowflakes from 
+         erasing already fallen snow a bit better
+ 30OCT94 More efficiency due to GC's in server and not drawing where
+         the windows are.
+ 30OCT94 1.30 Eiichi TAZOE's code for wind included in general distribution.
+ 11NOV94 Bug in XRectInRegion call resolved (santa not appearing gradually
+         from behind windows)
+ 15NOV94 Unsigned longs changed to longs, as strtoul is troublesome on many
+         systems (snowDelay)
+ 08DEC94 Snow is really building up on windows now
+ 10DEC94 1.36 Handle Popups (no snow on them)
+ 16DEC94 1.37 -offset and (-nokeepsnowonwindows and -nokeepsnowonscreen back)
+ 20DEC94 1.38 Let Santa be a real snowplough now
+ 21DEC94 1.39 VMS fix for uSsleep, thanks to Ulrich Kabatek 
+              <Kaba@isr.uni-stuttgart.d400.de>
 
 */
 
-#define VERSION "Xsnow-1.25, December 22 1993 by Rick Jansen (rick@sara.nl)"
+#define debug 0
+
+#define VERSION "Xsnow-1.39, December 21th 1994 by Rick Jansen (rick@sara.nl)"
+
+#ifdef VMS
+#include <socket.h>
+# if defined(__SOCKET_TYPEDEFS) && !defined(CADDR_T)
+#  define CADDR_T
+# endif
+#endif
 
 
+#include <X11/Intrinsic.h>
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/Xos.h>
@@ -87,11 +142,10 @@ Copyright 1984, 1988, 1990, 1993 by Rick Jansen, all rights reserved.
 
 #include <stdio.h>
 #include <math.h>
-#include <malloc.h>
 #include <signal.h>
 #include <limits.h>
 
-#if __STDC__
+#if defined(__STDC__) || defined(VMS)
 #include <stdlib.h>
 #else
 long strtol();
@@ -99,11 +153,11 @@ long strtol();
 char *getenv();
 #endif
 
-char Copyright[] = "Xsnow\nCopyright 1984, 1988, 1990, 1993 by Rick Jansen, all rights reserved.";
+char Copyright[] = "\nXsnow\nCopyright 1984, 1988, 1990, 1993, 1994 by Rick Jansen, all rights reserved.\n";
 
 #include "snowmap.h"
 
-typedef unsigned long Pixel;
+/***typedef unsigned long Pixel;***/
 typedef int ErrorHandler();
 
 #if !defined(GRAB_SERVER)
@@ -131,14 +185,16 @@ int Rudolf = 1;
 int NoKeepSnow = 0;
 int NoKeepSBot = 0;
 int NoKeepSWin = 0;
+int NoPopuphandling = 0;
+int NoWind = 0;
 unsigned int borderWidth = 0;
+int SnowOffset = 0;
 
 Snow *snowflakes;
 int MaxSnowFlakeHeight = 0;  /* Biggest flake */
 int maxSnowflakes = INITIALSNOWFLAKES;
 int curSnowflakes = 0;
-unsigned long snowDelay = 50000; /* useconds */
-unsigned long SnowTicks = ULONG_MAX; /* Decremented each pass thru main loop */
+long snowDelay = 50000; /* useconds */
 int MaxXStep = MAXXSTEP;
 int MaxYStep = MAXYSTEP;
 int WhirlFactor = WHIRLFACTOR;
@@ -151,17 +207,28 @@ Santa Claus;
 int SantaSize = 1;
 int SantaSpeed = -1;  /* Not initialized yet */
 
-Region SubR;
-Region WindowTops = NULL;
+
+/* For building up snow at bottom of screen and windows */
+Region Windows = NULL;
+Region SnowCatchR = NULL;
+Region SnowAllowedR = NULL;
+Region SubR = NULL;
+XRectangle AddRect;
+int AddX,AddY;
 int ClearX, ClearY;
 
-/* For building up snow at bottom of screen */
-Region SnowBottomR = NULL;
-XRectangle AddRect;
+/* Wind stuff */
+int diff=0;
+int wind = 0;
+int direction=0;
+int WindTimer=30;
+int current_snow_height;
+int geta=0;
 
 /* Forward decls */
 void Usage();
 void SigHandler();
+void SigHupHandler();
 void InitSnowflake();
 void UpdateSnowflake();
 void DrawSnowflake();
@@ -172,9 +239,10 @@ void InitSanta();
 void DrawSanta();
 void EraseSanta();
 void UpdateSanta();
-void usleep();
+void uSsleep();
 int CalcWindowTops();
 Pixel AllocNamedColor();
+void sig_alarm();
 
 /* Colo(u)rs */
 char *snowColor = "snow";
@@ -196,7 +264,17 @@ Pixel bgPix;
 Pixel trPix;
 Pixel slcPix;
 
+/* GC's */
+GC SnowGC[SNOWFLAKEMAXTYPE+1];
+GC SleighGCs[3];
+GC SantaGC,RudolfGC,FurGC;
+GC TreesGC;
+
+#ifdef VMS
+int
+#else
 void
+#endif
 main(ac, av)
 int ac;
 char *av[];
@@ -211,7 +289,11 @@ char *av[];
     SantaFurMap *stfp;
     XEvent ev;
     int needCalc;
+    int Exposed;
+    int ConfigureNotified;
     int i; 
+    int x,y;
+
     
     /*
        Process command line options.
@@ -234,7 +316,7 @@ char *av[];
 	    slcColor = av[++ax];
 	}
 	else if (strcmp(arg, "-delay") == 0) {
-	    snowDelay = strtod(av[++ax], (char **)NULL);
+	    snowDelay = strtol(av[++ax], (char **)NULL,0);
             snowDelay = snowDelay * 1000;  /* ms to us */
 	}
 	else if (strcmp(arg, "-snowflakes") == 0) {
@@ -246,6 +328,12 @@ char *av[];
 	else if (strcmp(arg, "-whirl") == 0) {
 	    WhirlFactor = strtol(av[++ax], (char **)NULL, 0);
 	}
+	else if (strcmp(arg, "-nowind") == 0) {
+	    NoWind = 1;
+	}
+	else if (strcmp(arg, "-windtimer") == 0) {
+	    WindTimer = strtol(av[++ax], (char **)NULL, 0);
+	}
 	else if (strcmp(arg, "-xspeed") == 0) {
 	    MaxXStep = strtol(av[++ax], (char **)NULL, 0);
 	}
@@ -258,6 +346,9 @@ char *av[];
 	else if (strcmp(arg, "-ssnowdepth") == 0) {
 	    MaxScrSnowDepth = strtol(av[++ax], (char **)NULL, 0);
 	}
+	else if (strcmp(arg, "-offset") == 0) {
+	    SnowOffset = strtol(av[++ax], (char **)NULL, 0);
+	}
         else if (strcmp(arg, "-notrees") == 0) {
             NoTrees = 1;     
         }
@@ -284,14 +375,20 @@ char *av[];
         else if (strcmp(arg, "-nokeepsnowonscreen") == 0) {
             NoKeepSBot = 1;
         }
+        else if (strcmp(arg, "-nonopopup") == 0) {
+            NoPopuphandling = 1;
+        }
         else if (strcmp(arg, "-version") == 0) {
             printf("\nThis is %s\n\n", VERSION);
+            exit(0);
         }
 	else {
 	    Usage();
 	}
     }
 
+    printf("%s\n",VERSION);
+
     /* Some people... */
     if (maxSnowflakes > MAXSNOWFLAKES) {
       fprintf(stderr,"xx Maximum number of snowflakes is %d\n", MAXSNOWFLAKES);
@@ -311,15 +408,29 @@ char *av[];
     if (SantaSpeed < 0) SantaSpeed = Speed[SantaSize];
 
     /* Seed random */
+#ifdef VMS
+    srand((int)time((unsigned long *)NULL));
+#else
     srand((int)time((long *)NULL));
-    
+#endif
+
+
     /*
        Catch some signals so we can erase any visible snowflakes.
     */
     signal(SIGKILL, SigHandler);
     signal(SIGINT, SigHandler);
     signal(SIGTERM, SigHandler);
+#if debug
+    signal(SIGHUP, SigHupHandler);
+#else
     signal(SIGHUP, SigHandler);
+#endif
+
+    /* Wind stuff */
+    signal(SIGALRM, sig_alarm);
+    alarm(WindTimer);
+
 
     /* Open X */
     display = XOpenDisplay(display_name);
@@ -340,6 +451,8 @@ char *av[];
     center_x = display_width / 2;
     center_y = display_height / 2;
 
+    /* No snow at all yet */
+    current_snow_height = display_height;
 
     /* Do not let all of the display snow under */
     if (MaxScrSnowDepth> (display_height-SNOWFREE)) {
@@ -410,19 +523,69 @@ char *av[];
       XFlush(display);
     }
     
+    /* Fill in GCs for snowflakes */
+    for (flake=0; flake<=SNOWFLAKEMAXTYPE; flake++) {
+      SnowGC[flake] = XCreateGC(display, rootWin, 0L, &xgcv);
+      XCopyGC(display,gc,0,SnowGC[flake]);
+      XSetStipple(display, SnowGC[flake], snowPix[flake].pixmap);
+      XSetForeground(display,SnowGC[flake],snowcPix);
+      XSetFillStyle(display, SnowGC[flake], FillStippled);
+    }
+    
+    /* GC for trees */
+    TreesGC = XCreateGC(display, rootWin, 0L, &xgcv);
+    XCopyGC(display,gc,0,TreesGC);
+    XSetStipple(display, TreesGC, tannenbaumPix[0].pixmap);
+    XSetForeground(display,TreesGC,trPix);
+    XSetFillStyle(display, TreesGC, FillStippled);
+
+    /* GCs for sleighs */
+    for (flake=0; flake<=2; flake++) {
+      SleighGCs[flake] = XCreateGC(display, rootWin, 0L, &xgcv);
+      XCopyGC(display,gc,0,SleighGCs[flake]);
+      XSetForeground(display, SleighGCs[flake], slcPix);
+      XSetFillStyle(display, SleighGCs[flake], FillStippled);
+      XSetStipple(display,SleighGCs[flake],
+                  sleighPix[SantaSize][flake].pixmap);
+    }
+
+    /* GCs for Santa */
+    SantaGC = XCreateGC(display, rootWin, 0L, &xgcv);
+    XCopyGC(display,gc,0,SantaGC);
+    XSetFillStyle(display, SantaGC, FillStippled);
+    XSetForeground(display, SantaGC, redPix);
+    XSetStipple(display,SantaGC,
+                santaPix[SantaSize].pixmap);
+
+    RudolfGC = XCreateGC(display, rootWin, 0L, &xgcv);
+    XCopyGC(display,gc,0,RudolfGC);
+    XSetFillStyle(display, RudolfGC, FillSolid);
+    XSetForeground(display, RudolfGC, redPix);
+
+    FurGC = XCreateGC(display, rootWin, 0L, &xgcv);
+    XCopyGC(display,gc,0,FurGC);
+    XSetForeground(display, FurGC, whitePix);
+    XSetFillStyle(display, FurGC, FillStippled);
+    XSetStipple(display, FurGC, santaFurPix[SantaSize].pixmap);
+
+
     /*
      *
      * Here snowing starts
      *
      */
 
+    Windows = XCreateRegion();
+    SnowCatchR = XCreateRegion();
+    SnowAllowedR = XCreateRegion();
+
     /* The initial catch region: a little bit below the bottom of the screen */
     AddRect.x = 0;
-    AddRect.y = display_height + MaxSnowFlakeHeight/2;
+    AddRect.y = display_height;
     AddRect.width = display_width - 1;
     AddRect.height = MaxYStep+MaxSnowFlakeHeight;
-    SnowBottomR = XCreateRegion();
-    XUnionRectWithRegion(&AddRect, SnowBottomR, SnowBottomR);
+    if (!NoKeepSBot)
+      XUnionRectWithRegion(&AddRect, SnowCatchR, SnowCatchR);
 
     /* Initialize all snowflakes */
     for (i=0; i<maxSnowflakes; i++) InitSnowflake(i);
@@ -432,8 +595,10 @@ char *av[];
     /* Notify if part of the root window changed */
     XSelectInput(display, rootWin, ExposureMask | SubstructureNotifyMask);
 
+
     needCalc = 0;
-    if (!NoKeepSnow) needCalc = CalcWindowTops();
+    /***if (!NoKeepSnow) needCalc = CalcWindowTops();***/
+    needCalc = CalcWindowTops();
 
 
     /*
@@ -441,11 +606,10 @@ char *av[];
      */
     while (!done) {
 
-      SnowTicks--;   
-      if (SnowTicks == 0) SnowTicks = ULONG_MAX;
-
       /* X event ? */
       /* Handle all the expose events and redo CalcWindowTops after */
+      Exposed = 0;
+      ConfigureNotified = 0;
       while (XPending(display)) {
         XNextEvent(display, &ev);
 
@@ -455,29 +619,49 @@ char *av[];
           switch (ev.type) {
 
               case Expose:    
+                
+                  Exposed = 1;
 
-                  /* Subtract the exposed area from the SnowBottomR */
-                  AddRect.x = ev.xexpose.x;
-                  AddRect.y = ev.xexpose.y;
-                  AddRect.width = ev.xexpose.width;
-                  /* Add MSFH so snow will really build up from scratch */
-                  AddRect.height = ev.xexpose.height + MaxSnowFlakeHeight/2;
+                  /* 
+                   * Subtract the exposed area from the SnowCatchR 
+                   */
+                  AddX = ev.xexpose.x - MaxSnowFlakeHeight;
+                  AddY = ev.xexpose.y - MaxWinSnowDepth;
+                  AddRect.height = ev.xexpose.height +  MaxWinSnowDepth;
+                  AddRect.width = ev.xexpose.width + 2*MaxSnowFlakeHeight;
+                  if (AddY < 0) {
+                    AddRect.height = AddRect.height + AddY;
+                    AddY = 0;
+                  }
+                  if ((AddY+AddRect.height) > display_height) {
+                    AddRect.height = display_height - AddY;
+                  }
+
+                  AddRect.x = AddX;
+                  AddRect.y = AddY;
 
                   SubR= XCreateRegion();
-                  XUnionRectWithRegion(&AddRect, SubR,SubR);
-                  XSubtractRegion(SnowBottomR,SubR, SnowBottomR);
-                  XDestroyRegion(SubR);
+                    XUnionRectWithRegion(&AddRect, SubR,SubR);
                   
-                  /* Clear the snow that was on top of the window */
-                  ClearX = ev.xexpose.x - MaxWinSnowDepth - MaxXStep;
-                  if (ClearX < 0) ClearX = 0;
-                  ClearY = ev.xexpose.y-MaxWinSnowDepth-MaxYStep-borderWidth;
-                  if (ClearY < 0) ClearY = 0;
+                    /* Subtract the erased snow from the SnowCatchR */
+                    XSubtractRegion(SnowCatchR,SubR, SnowCatchR);
+                  XDestroyRegion(SubR);
+
+                  /* Clear the snow from screen */
+                  AddY = AddY - MaxYStep-MaxSnowFlakeHeight;
+                  AddRect.height = 
+                    MaxWinSnowDepth+MaxYStep*2+MaxSnowFlakeHeight*2;
+                  if (AddY < 0) AddY = 0;
                   XClearArea(display, rootWin,
-                     ClearX,ClearY, 
-                     ev.xexpose.width+ 2*(MaxWinSnowDepth+MaxYStep),
-                     MaxWinSnowDepth+MaxYStep+borderWidth,
-                     False);
+                             AddX,
+                             AddY,
+                             AddRect.width,              
+                             AddRect.height,
+                             False);
+
+#if debug
+printf("exp %d %d %d %d\n",AddX,AddY,AddRect.width,AddRect.height);
+#endif
 
                   needCalc = 1;
 
@@ -487,6 +671,9 @@ char *av[];
               case MapNotify:
               case UnmapNotify:
               case ConfigureNotify:
+#if debug
+printf("Configurenotify!\n");
+#endif
                   needCalc = 1;
                   break;
 
@@ -500,16 +687,31 @@ char *av[];
       if (needCalc) needCalc = CalcWindowTops();
 
 
+#if debug
+/***
+XClearWindow(display, rootWin);
+XFlush(display);
+XSetForeground(display,FurGC,trPix);
+XSetFillStyle(display,FurGC, FillSolid);
+XSetRegion(display,FurGC,SnowCatchR);
+XFillRectangle(display,rootWin,FurGC,0,0,display_width,display_height);
+XFlush(display);
+***/
+#endif
+
+
+      /* Sleep a bit */
+      uSsleep(snowDelay);
+
       /* 
        *  Update     
        */
 
-      XSetForeground(display, gc, snowcPix);
+      /* Snowflakes */
       for (i=0; i<maxSnowflakes; i++) UpdateSnowflake(i);
 
       /* Draw dear Andrews */
       if (!NoTrees ) {
-        XSetForeground(display, gc, trPix);
         DrawTannenbaum(display_width-150, display_height-500);
         DrawTannenbaum(display_width-100, display_height-200);
         DrawTannenbaum(100, display_height-200);
@@ -520,15 +722,25 @@ char *av[];
 
       /* Dear Santa */
       if (!NoSanta) {
-        XSetForeground(display, gc, slcPix);
         UpdateSanta();
       }
 
-      /* Sleep a bit */
-      usleep(snowDelay);
     }
+
+/***
+sleep(10);
+XClearWindow(display, rootWin);
+XFlush(display);
+XSetFillStyle(display,FurGC, FillSolid);
+XSetRegion(display,FurGC,SnowAllowedR);
+XFillRectangle(display,rootWin,FurGC,0,0,display_width,display_height);
+XFlush(display);
+sleep(4);
+***/
     
-    if (!NoKeepSnow) XDestroyRegion(WindowTops);
+    XDestroyRegion(Windows);
+    XDestroyRegion(SnowCatchR);
+    XDestroyRegion(SnowAllowedR);
 
     XClearWindow(display, rootWin);
 
@@ -554,9 +766,12 @@ Usage()
     USEPRT("       -delay       <delay in milliseconds>\n");
     USEPRT("       -unsmooth\n");
     USEPRT("       -whirl       <whirlfactor>\n");
-    USEPRT("       -xspeed      <max xspeed>\n");
-    USEPRT("       -yspeed      <max yspeed>\n");
+    USEPRT("       -nowind\n");
+    USEPRT("       -windtimer   <Time between windy periods in seconds>\n");
+    USEPRT("       -xspeed      <max xspeed snowflakes>\n");
+    USEPRT("       -yspeed      <max yspeed snowflakes>\n");
     USEPRT("       -wsnowdepth  <max snow depth on windows>\n");
+    USEPRT("       -offset      <shift snow down>\n");
     USEPRT("       -ssnowdepth  <max snow depth at bottom of display>\n");
     USEPRT("       -notrees\n");
     USEPRT("       -nosanta\n");
@@ -566,6 +781,7 @@ Usage()
     USEPRT("       -nokeepsnow\n");
     USEPRT("       -nokeepsnowonwindows\n");
     USEPRT("       -nokeepsnowonscreen\n");
+    USEPRT("       -nonopopup\n");
     USEPRT("       -version\n\n");
     USEPRT("Recommended: xsnow -bg LightSkyBlue3\n");
     
@@ -578,6 +794,23 @@ SigHandler()
   done = 1;
 }
 
+
+void
+SigHupHandler()
+{
+  void (*Sig_Hup_ptr)() = SigHupHandler;
+
+  /*XClearWindow(display, rootWin);*/
+  XFlush(display);
+  XSetForeground(display,FurGC,trPix);
+  XSetFillStyle(display,FurGC, FillSolid);
+  XSetRegion(display,FurGC,SnowCatchR);
+  XFillRectangle(display,rootWin,FurGC,0,0,display_width,display_height);
+  XFlush(display);
+
+  signal(SIGHUP, Sig_Hup_ptr);
+}
+
 /*
    Generate random integer between 0 and maxVal-1.
 */
@@ -594,21 +827,27 @@ int maxVal;
 /*
  * sleep for a number of micro-seconds
  */
-void usleep(usec) 
+void uSsleep(usec) 
 unsigned long usec;
 {
 #ifdef SVR3
     poll((struct poll *)0, (size_t)0, usec/1000);   /* ms resolution */
+#else
+#ifdef VMS
+    void lib$wait();
+    float t;
+    t= usec/1000000.0;
+    lib$wait(&t);
 #else
     struct timeval t;
     t.tv_usec = usec%(unsigned long)1000000;
     t.tv_sec = usec/(unsigned long)1000000;
     select(0, (void *)0, (void *)0, (void *)0, &t);
 #endif
+#endif
 }
 
 
-
 /*
    Check for topleft point of snowflake in specified rectangle.
 */
@@ -643,16 +882,33 @@ int rx;
 
   if (curSnowflakes < maxSnowflakes) {
     r->whatFlake = RandInt(SNOWFLAKEMAXTYPE+1);
-    r->intX = RandInt(display_width - snowPix[r->whatFlake].width);
-    r->intY =  RandInt(display_height/15);
+
+    /* Wind strikes! */
+    if (wind) {
+      if (direction == 1) 
+        /* Create snow on the left of the display */
+        r->intX = RandInt(display_width/3);
+      else
+        /* Create snow on the right of the display */
+        r->intX = display_width - RandInt(display_width/3);
+      r->intY =  RandInt(display_height);
+    }
+    else  {
+      r->intX = RandInt(display_width - snowPix[r->whatFlake].width);
+      r->intY =  RandInt(display_height/10);
+    }
+
     r->yStep = RandInt(MaxYStep+1)+1;
     r->xStep = RandInt(r->yStep/4+1);
     if (RandInt(1000) > 500) r->xStep = -r->xStep;
     r->active = 1;
+    r->visible = 1;
+    r->insnow = 0;
   }
 }
 
 
+
 /*
    Move a snowflake by erasing and redraw
 */
@@ -664,7 +920,9 @@ Snow *snow;
 
 int NewX;
 int NewY;
+int tmp_x;
 int TouchDown;
+int InVisible;
 int NoErase;
     
   /* move an active snowflake */
@@ -674,90 +932,141 @@ int NoErase;
 
 
   /* Activate a new snowflake */
-  if (!snow->active) InitSnowflake(rx);
+  if (!snow->active) {
+    InitSnowflake(rx);
+    /* Make sure newly created flakes aren't erased before actually drawn */
+    /* Especially with wind, where flakes are created all over the screen */
+    snow->insnow = 1; 
+  }
 
+  /* New x,y */
+  if (wind) {
+    tmp_x = abs(snow->xStep);
+    if( wind == 1 ){  /* going to stop winding */
+        tmp_x = tmp_x + (RandInt(WhirlFactor+1) - WhirlFactor/2);
+    }else{            /* WINDY */
+        tmp_x = tmp_x + (RandInt(20));
+    }
+    snow->xStep = tmp_x * direction;
 
+    if (snow->xStep > 50) snow->xStep = 50;
+    if (snow->xStep < -50) snow->xStep = -50;
+
+    if( wind == 1 ){
+        NewY = NewY+3;
+    }else{
+        NewY = NewY+10;
+    }
+  }
   NewX = snow->intX + snow->xStep;
   NewY = snow->intY + snow->yStep;
 
-  /* If flake disappears offscreen it becomes inactive */
-  snow->active = SnowPtInRect(NewX, NewY,
-                              0, 0, display_width-1, display_height);
+
+  /* If flake disappears offscreen (below bottom) it becomes inactive */
+  /* In case of wind also deactivate flakes blown offscreen sideways, */
+  /* so they'll reappear immediately. Make sure snow at bottom isn't  */
+  /* disrupted.                          */
+
+  /* Snowflake at the end of it's life? */
+  snow->active = (NewY < display_height);
+  /* If it is windy we need the flakes that blow offscreen! */
+  if (wind) snow->active = 
+      (snow->active && (NewX > 0) && (NewX < display_width));
+    
+
+  /* If flake covered by windows don't bother drawing it */
+  InVisible = XRectInRegion(Windows,
+                            NewX,NewY,
+                            snowPix[snow->whatFlake].width,
+                            snowPix[snow->whatFlake].height);
+  snow->visible = (InVisible != RectangleIn);
+
 
   /*
-   * Screen bottom 
+   *  Snow touches snow
    */
-  /* Keep snow that becomes inactive */
-  if (!NoKeepSBot) {
+  if (snow->visible) {
 
-    TouchDown = 
-       /* Note: if we are too strict and use intXY the snow won't be fluffy */
-       /* We really need some flakes just hanging in the air.               */
-       /* Correction! We mean, sticking to the *glass*, of course. Ehem.    */
-       XRectInRegion(SnowBottomR,
-                     NewX,NewY,
-                     snowPix[snow->whatFlake].height,
-                     snowPix[snow->whatFlake].width);
+    /* Do not use XRectInRegion here, as that makes snow add to the bottom */
+    /* of layers on windows too */
+    /* Use middle bottom of flake */
+    TouchDown = XPointInRegion(SnowCatchR,
+                               NewX+snowPix[snow->whatFlake].width/2,
+                               NewY+snowPix[snow->whatFlake].height);
 
     /* If the flake already is completely inside the region */
-    if (TouchDown == RectangleIn) 
-      NoErase = 1;
-    else
-    if (TouchDown == RectanglePart) {
+    if (TouchDown && snow->visible && !NoKeepSnow) {
+      /* Only add snow that's visible, else snow will grow */
+      /* on the *back* of windows */
 
       NoErase = 1;
 
       /* Add a bit to the snow bottom region to make snow build up */
-      /* Snow isn't built up more than about MaxScrSnowDepth */
-      AddRect.y = NewY + snowPix[snow->whatFlake].height - 2;
-      /* Limit the building-up */
-      if (AddRect.y > (display_height-MaxScrSnowDepth)) {
+      /* Snow isn't built up more than about a certain amount: inside */
+      /* SnowAllowedR */
+
+      /* If flake already inside snow layer don't add it again */
+      TouchDown = XRectInRegion(SnowCatchR,
+                                NewX,NewY,
+                                snowPix[snow->whatFlake].width,
+                                snowPix[snow->whatFlake].height);
+      if (TouchDown == RectanglePart) {
         AddRect.x = NewX;
+        AddRect.y = NewY + snowPix[snow->whatFlake].height - 2;
         AddRect.height = 2;
         AddRect.width = snowPix[snow->whatFlake].width;               
-        XUnionRectWithRegion(&AddRect, SnowBottomR,SnowBottomR);
+        if (XRectInRegion(SnowAllowedR,
+                          AddRect.x, AddRect.y,
+                          AddRect.width, AddRect.height) == RectangleIn) {
+          XUnionRectWithRegion(&AddRect, SnowCatchR,SnowCatchR);
+        }
       }
     }
 
-  }  /* if !NoKeepSBot */
+
+    /* Adjust horizontal speed to mimic whirling */
+    if (SmoothWhirl) 
+      snow->xStep = snow->xStep + (RandInt(WhirlFactor+1) - WhirlFactor/2);
+    else {
+      /* Jerkier movement */
+      snow->xStep = snow->xStep + RandInt(WhirlFactor);
+      if (RandInt(1000) > 500) snow->xStep = -snow->xStep;
+    }
+
+  }  /* if snow->visible */
 
 
-  /*
-   * Window tops 
-   */
-  /* Are we keeping snow on top of the windows? */
-  /* Note snow kept on top of a window is still active, so it 
-     can reach the bottom of the screen... */
-  if (snow->active) {
-    if (!NoKeepSWin) 
-      NoErase = (NoErase || XPointInRegion(WindowTops,NewX,NewY));
+
+  if (!wind) {
+    if (snow->xStep > MaxXStep) snow->xStep = MaxXStep;
+    if (snow->xStep < -MaxXStep) snow->xStep = -MaxXStep;
   }
 
-  /* Adjust horizontal speed to mimic whirling */
-  if (SmoothWhirl) 
-    snow->xStep = snow->xStep + (RandInt(WhirlFactor+1) - WhirlFactor/2);
-  else {
-    /* Jerkier movement */
-    snow->xStep = snow->xStep + RandInt(WhirlFactor);
-    if (RandInt(1000) > 500) snow->xStep = -snow->xStep;
-  }
-
-  if (snow->xStep > MaxXStep) snow->xStep = MaxXStep;
-  if (snow->xStep < -MaxXStep) snow->xStep = -MaxXStep;
-
   /* Limit X speed inside snowlayers. 's only natural... */
-  if (NoErase) snow->xStep = snow->xStep/2;
+  if (NoErase && !wind) snow->xStep = snow->xStep/2;
+  /* xsnow 1.29 method: */
+  /***if (NoErase) {
+    snow->xStep = 0;
+    NewX = snow->intX;
+  }***/
+
 
   /* Erase, unless something special has happened */
-  if (!NoErase) EraseSnowflake(rx);
+  if (!snow->insnow) EraseSnowflake(rx);
 
-  /* Draw if still active */
+  /* New coords, unless it's windy */
   snow->intX = NewX;
   snow->intY = NewY;
-  if (snow->active) DrawSnowflake(rx);
+
+  /* If snowflake inside a snow layer now don't erase it next time */
+  snow->insnow = NoErase;
+
+  /* Draw if still active  and visible*/
+  if (snow->active && snow->visible) DrawSnowflake(rx);
  
 }
 
+
     
 /*
    Draw a snowflake.
@@ -770,9 +1079,8 @@ int rx;
 
     snow = &snowflakes[rx];
 
-    XSetStipple(display, gc, snowPix[snow->whatFlake].pixmap);
-    XSetTSOrigin(display, gc, snow->intX, snow->intY);
-    XFillRectangle(display, rootWin, gc,
+    XSetTSOrigin(display, SnowGC[snow->whatFlake], snow->intX, snow->intY);
+    XFillRectangle(display, rootWin, SnowGC[snow->whatFlake],
          snow->intX, snow->intY,
          snowPix[snow->whatFlake].width, snowPix[snow->whatFlake].height);
 }
@@ -807,8 +1115,6 @@ int depth;
 int x,y;
 Snow *snow;
 
-  XSetForeground(display, gc, snowcPix);
-
   for (y=0; y < depth; y++) {
     for (x=0; x<maxSnowflakes; x++) {
       InitSnowflake(x);
@@ -828,9 +1134,8 @@ void
 DrawTannenbaum(x,y)
 int x,y;
 {
-    XSetStipple(display, gc, tannenbaumPix[0].pixmap);
-    XSetTSOrigin(display, gc, x,y);
-    XFillRectangle(display, rootWin, gc,
+    XSetTSOrigin(display, TreesGC, x,y);
+    XFillRectangle(display, rootWin, TreesGC,
          x,y,
          tannenbaumPix[0].width, tannenbaumPix[0].height);
 }
@@ -842,11 +1147,12 @@ int x,y;
 void
 InitSanta()      
 {
-  Claus.x = 0;
+  Claus.x = -sleighPix[SantaSize][Claus.whatSanta].width;
   Claus.y = RandInt(display_height / 3)+40;
-  Claus.xStep = SantaSpeed;;
+  Claus.xStep = SantaSpeed;
   Claus.yStep = 1;
   Claus.whatSanta = 0;
+  Claus.visible = 1;
 }
 
 
@@ -856,7 +1162,9 @@ InitSanta()
 void
 UpdateSanta()
 {
-  EraseSanta();
+int Visible;
+
+  if (Claus.visible) EraseSanta();
 
   /* Move forward */
   Claus.x = Claus.x + Claus.xStep;
@@ -868,11 +1176,43 @@ UpdateSanta()
 
   if (Claus.x >= display_width) InitSanta(); 
 
+  /* If Santa entirely covered by windows don't bother drawing him */
+  Visible = XRectInRegion(Windows,
+                          Claus.x, Claus.y,
+                          sleighPix[SantaSize][Claus.whatSanta].width,
+                          sleighPix[SantaSize][Claus.whatSanta].height);
+  Claus.visible = (Visible != RectangleIn);
+  Claus.visible = (Claus.visible || (Claus.x < 0));
+
+  /* Santa inside snow layers? Then make a hole in the snow :)  */
+  Visible = XRectInRegion(SnowCatchR,
+                          Claus.x, Claus.y,
+                          sleighPix[SantaSize][Claus.whatSanta].width,
+                          sleighPix[SantaSize][Claus.whatSanta].height);
+
+  /* Santa acts as a real snowplough! */
+  if (Visible != RectangleOut) {
+    SubR = XCreateRegion();
+
+    AddRect.x = Claus.x;
+    AddRect.y = Claus.y;
+    AddRect.width = sleighPix[SantaSize][Claus.whatSanta].width;
+    AddRect.height = sleighPix[SantaSize][Claus.whatSanta].height;
+    XUnionRectWithRegion(&AddRect, SubR,SubR);
+    XSubtractRegion(SubR,Windows, SubR);
+ 
+    XSubtractRegion(SnowCatchR,SubR, SnowCatchR);
+
+    XDestroyRegion(SubR);
+  }
+
+  
+
   /* Next sleigh */
   Claus.whatSanta++;
   if (Claus.whatSanta > 2) Claus.whatSanta = 0;
 
-  DrawSanta();
+  if (Claus.visible) DrawSanta();
 }
 
 
@@ -884,39 +1224,34 @@ UpdateSanta()
 void
 DrawSanta()
 {
-  XSetStipple(display, gc, sleighPix[SantaSize][Claus.whatSanta].pixmap);
-  XSetTSOrigin(display, gc, Claus.x,Claus.y);
-  XFillRectangle(display, rootWin, gc,
+  XSetTSOrigin(display, SleighGCs[Claus.whatSanta], Claus.x,Claus.y);
+  XFillRectangle(display, rootWin, SleighGCs[Claus.whatSanta],
          Claus.x,Claus.y,
          sleighPix[SantaSize][Claus.whatSanta].width, 
          sleighPix[SantaSize][Claus.whatSanta].height);
 
   /* The Man in the Red Suit himself */
-  XSetForeground(display, gc, redPix);
-  XSetStipple(display, gc, santaPix[SantaSize].pixmap);
-  XSetTSOrigin(display, gc, Claus.x,Claus.y);
-  XFillRectangle(display, rootWin, gc,
+  XSetTSOrigin(display, SantaGC, Claus.x,Claus.y);
+  XFillRectangle(display, rootWin, SantaGC,
          Claus.x,Claus.y,
-         santaPix[SantaSize].width, santaPix[SantaSize].height);
+         santaPix[SantaSize].width, 
+         santaPix[SantaSize].height);
 
   /* At last.... Rudolf! */
   if (Rudolf) {
-    XSetFillStyle(display, gc, FillSolid);
-    XFillRectangle(display, rootWin, gc,
+    XFillRectangle(display, rootWin, RudolfGC,
                    Claus.x+NoseX[SantaSize], Claus.y+NoseY[SantaSize],
                    NoseXSiz[SantaSize], NoseYSiz[SantaSize]);
-    XSetFillStyle(display, gc, FillStippled);
   }
 
   /* The fur in his hat */
   /* Note: the fur in his hat is *imitation* white-seal fur, of course. */
   /* Santa is a big supporter of Greenpeace.                            */
-  XSetForeground(display, gc, whitePix);
-  XSetStipple(display, gc, santaFurPix[SantaSize].pixmap);
-  XSetTSOrigin(display, gc, Claus.x,Claus.y);
-  XFillRectangle(display, rootWin, gc,
+  XSetTSOrigin(display, FurGC, Claus.x,Claus.y);
+  XFillRectangle(display, rootWin, FurGC,
          Claus.x,Claus.y,
-         santaFurPix[SantaSize].width, santaFurPix[SantaSize].height);
+         santaFurPix[SantaSize].width, 
+         santaFurPix[SantaSize].height);
 
 }
 
@@ -926,12 +1261,11 @@ DrawSanta()
 void
 EraseSanta()
 {
-  if (Claus.x >= 0) 
-      XClearArea(display, rootWin,
-                 Claus.x , Claus.y,     
-                 sleighPix[SantaSize][Claus.whatSanta].width,
-                 sleighPix[SantaSize][Claus.whatSanta].height,
-                 False);
+  XClearArea(display, rootWin,
+             Claus.x , Claus.y,     
+             sleighPix[SantaSize][Claus.whatSanta].width,
+             sleighPix[SantaSize][Claus.whatSanta].height,
+             False);
 }
 
 
@@ -956,14 +1290,16 @@ XErrorEvent *err;
 int
 CalcWindowTops()
 {
-    Region covered;
     Window *children;
     unsigned int nChildren;
     Window dummy;
     XWindowAttributes wa;
     int wx;
-    XRectangle rect;
+    XRectangle CatchRect;
+    XRectangle AllowRect;
+    XRectangle WinRect;
     int winX, winY;
+    int NouMoe;
     unsigned int winHeight, winWidth;
     unsigned int depth;
 
@@ -978,6 +1314,12 @@ CalcWindowTops()
     XSetErrorHandler(RoachErrors);
 #endif
 
+    /* Rebuild regions */
+    XDestroyRegion(Windows);
+    XDestroyRegion(SnowAllowedR);
+    Windows = XCreateRegion();
+    SnowAllowedR = XCreateRegion();
+
     /*
        Get children of root.
     */
@@ -987,37 +1329,119 @@ CalcWindowTops()
        For each mapped child, add the rectangle on top of the window
        where the snow could be to the region.
     */
-    covered = XCreateRegion();
     for (wx=0; wx<nChildren; wx++) {
         if (XEventsQueued(display, QueuedAlready)) return 1;
         errorVal = 0;
         XGetWindowAttributes(display, children[wx], &wa);
         if (errorVal) continue;
+
+        /* Popup? */
+#if debug
+if (wa.save_under) printf("POPUP discarded\n");;
+#endif
+        if (!NoPopuphandling && wa.save_under) continue;
+
         if (wa.map_state == IsViewable) {
             XGetGeometry(display, children[wx], &dummy, &winX, &winY,
                 &winWidth, &winHeight, &borderWidth, &depth);
+#if debug
+printf("\nw %d %d %d %d - %d %d \n", winX,winY,winWidth,winHeight,borderWidth,depth);
+#endif
             if (errorVal) continue;
-            rect.x = winX - borderWidth - MaxWinSnowDepth;
-
-            /* Patch from Mike Hollick <hollick@graphics.cis.upenn.edu> */
-            /* for fvwm. 14DEC93                                        */
+ 
+winY = winY + SnowOffset;
             /* Entirely offscreen? */
-            if (((rect.x + winWidth) < 0) || (rect.x > display_width)) continue;
+            if (winX > display_width) continue;
+            if (winY > display_height) continue;
+            if (winY <= 0) continue;
+
+
+            /* Geometry of the window, borders inclusive */
+            winX = winX-borderWidth;
+            winY = winY-borderWidth;
+            winHeight = winHeight + 2*borderWidth;
+            winWidth = winWidth + 2*borderWidth;
+
+            /* Entirely offscreen? */
+            NouMoe = winX + winWidth;
+            if (NouMoe < 0) continue;
+
+            /* Make sure the initial catch region isn't punctured */
+            if ((winY+winHeight) > display_height) {
+              winHeight = display_height-winY;
+            }
+
+            if (winY < 0) {
+              winHeight = winHeight + winY;
+              winY = 0;
+            }
 
             /* If window partly on screen */
-            if (rect.x < 0) {
-              winWidth = winWidth + rect.x;
-              rect.x = 0;
+            if (winX < 0) {
+              winWidth = winWidth + winX - borderWidth;
+              winX = 0;
             }
-              
-            rect.y = winY - MaxWinSnowDepth - borderWidth;
-            if (rect.y < 0) continue;
 
-            rect.width = winWidth + MaxWinSnowDepth + borderWidth*2;
-            rect.height = MaxWinSnowDepth + borderWidth + MaxSnowFlakeHeight; 
-            XUnionRectWithRegion(&rect, covered, covered);
+            WinRect.x = winX;
+            WinRect.y = winY;
+            WinRect.height = winHeight + 2*borderWidth;
+            WinRect.width = winWidth + 2*borderWidth;
+
+            /* The area of the windows themselves */
+            XUnionRectWithRegion(&WinRect, Windows, Windows);
+
+
+
+
+
+            /* Subtract this window from all previous catch regions! */
+            SubR= XCreateRegion();
+              XUnionRectWithRegion(&WinRect, SubR, SubR);
+
+              XSubtractRegion(SnowCatchR,SubR, SnowCatchR);
+            XDestroyRegion(SubR);
+              
+
+
+            /* CatchR */
+            CatchRect.y = WinRect.y;
+            CatchRect.x = WinRect.x;
+            CatchRect.height = MaxYStep+MaxSnowFlakeHeight;
+            CatchRect.width = WinRect.width;
+
+            if (!NoKeepSWin) 
+              XUnionRectWithRegion(&CatchRect, SnowCatchR, SnowCatchR);
+#if debug
+printf("c %d %d %d %d\n",CatchRect.x,CatchRect.y,CatchRect.height,CatchRect.width);
+#endif
+
+
+            /* AllowedR */
+            AllowRect.x = CatchRect.x;
+            winY = CatchRect.y - MaxWinSnowDepth;
+            AllowRect.height = CatchRect.height+MaxWinSnowDepth;
+            if (winY < 0) {
+              AllowRect.height = winY+CatchRect.height;
+              winY = 0;
+            }
+            AllowRect.y = winY;
+            AllowRect.width = CatchRect.width;
+            if (!NoKeepSWin)
+              XUnionRectWithRegion(&AllowRect, SnowAllowedR, SnowAllowedR);
+
+
         }
     }
+
+    /* SnowAllowed at screen bottom */
+    AllowRect.x = 0;
+    AllowRect.y = display_height - MaxScrSnowDepth;
+    AllowRect.width = display_width - 1;
+    AllowRect.height = MaxYStep+MaxSnowFlakeHeight+MaxScrSnowDepth;
+    if (!NoKeepSWin)
+      XUnionRectWithRegion(&AllowRect, SnowAllowedR, SnowAllowedR);
+
+
     XFree((char *)children);
 
 #if GRAB_SERVER
@@ -1026,12 +1450,6 @@ CalcWindowTops()
     XSetErrorHandler((ErrorHandler *)NULL);
 #endif
 
-    /*
-       Save window tops globally.
-    */
-    if (WindowTops)
-        XDestroyRegion(WindowTops);
-    WindowTops = covered;
 
     return 0;
 }
@@ -1061,3 +1479,43 @@ Pixel dfltPix;
 
 	return pix;
 }
+
+
+void sig_alarm()
+{
+    int rand=RandInt(100);
+    void (*sig_alarm_ptr)() = sig_alarm;
+
+    /* No wind at all? */
+    if (NoWind) return;
+
+    if( rand > 80 ){
+        geta -= 1;
+    }else if( rand > 55 ){
+        geta += 1;
+    }
+
+    if( rand > 65 ){
+        if( RandInt(10) > 4 ){
+            direction = 1;
+        }else{
+            direction = -1;
+        }
+
+        wind = 2;
+        signal(SIGALRM, sig_alarm_ptr);
+
+        alarm(abs(RandInt(5))+1);
+    }else{
+        if( wind == 2 ){
+            wind = 1;
+            signal(SIGALRM, sig_alarm_ptr);
+            alarm(RandInt(3)+1);
+        }else{
+            wind = 0;
+            signal(SIGALRM, sig_alarm_ptr);
+            alarm(WindTimer);
+        }
+    }
+}
+
